#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>


#include <vm68k/types.h>
#include <vm68k/memory.h>
#include <vm68k/processor.h>


#include "ledoutmmdevice.hh"
#include "mmdevice.hh"
#include "programram.hh"
#include "ram.hh"
#include "ttyoutmmdevice.hh"

using namespace vm68k ;

static context * m68kContext ;

static void printRegs ( void ) ;

static void sigIntHandler ( int s )
{
    fprintf ( stderr , "^C\n" ) ;
    printRegs () ;
    exit ( 1 ) ;
}


int main ( int argc , char * argv [] )
{
    const uint32_type   memorySize = 32768 ;
    const char          file_name [] = "a.out" ;
    memory_map *        m68kMemoryMap ;
    exec_unit *         m68kExecution ;
    ProgramRAM *        m68kROMMemory ;
    RAM *               m68kRAMMemory ;
    TTYOutMMDevice *    m68kOutDev ;
    LEDOutMMDevice *    m68kLEDOutDev ;
    struct sigaction    sa ;

    m68kMemoryMap = new memory_map () ;
    m68kExecution = new exec_unit () ;

    try
    {
        m68kROMMemory = new ProgramRAM ( 0 , 0xfff , file_name ) ;
    } catch ( exception e )
    {
        fprintf ( stderr , "Unable to read file '%s'.\n" , file_name ) ;
        return 1 ;
    }

    m68kOutDev = new TTYOutMMDevice ( stderr ) ;
    m68kLEDOutDev = new LEDOutMMDevice ( stderr ) ;
    m68kRAMMemory = new RAM ( 0x1000 , 0x7fff ) ;

    m68kMemoryMap -> fill ( 0 , 0xfff , m68kROMMemory ) ;
    m68kMemoryMap -> fill ( 0x1000 , 0x7fff , m68kRAMMemory ) ;
    m68kMemoryMap -> fill ( 0x8004 , 0x8007 , m68kOutDev ) ;
    m68kMemoryMap -> fill ( 0x8000 , 0x8003 , m68kLEDOutDev ) ;

    m68kContext = new context ( m68kMemoryMap ) ;

    /* Initialise sigint handler */
    sa . sa_handler = sigIntHandler ;
    sigemptyset ( & sa . sa_mask ) ;
    sigaction ( SIGINT , & sa , NULL ) ;

    /* Clear all the registers */
    for ( int i = 0 ; i < 8 ; i ++ )
    {
        m68kContext -> regs . d [ i ] =
            m68kContext -> regs . a [ i ] = 0 ;
    }
    m68kContext -> regs . usp = m68kContext -> regs . ssp = 0 ;

    /* Except for these essential ones */
    m68kContext -> regs . pc = m68kROMMemory -> getStartPC () ;
    m68kContext -> regs . a [ 7 ] = memorySize - 2 ; /* stack ptr */

    fprintf ( stderr , "Program '%s' starts.\n\n" , file_name ) ;

    try
    {
        m68kExecution -> run ( (* m68kContext) ) ;
    } catch ( bus_error e )
    {
        fprintf ( stderr , "\n\nvm68k: Bus error: address 0x%08x %s\n" , 
                e . _address , ( e . _status & memory::READ ) ? "RD" : "WR" ) ;
    } catch ( address_error e )
    {
        fprintf ( stderr , "\n\nvm68k: Address error: address 0x%08x %s\n" , 
                e . _address , ( e . _status & memory::READ ) ? "RD" : "WR" ) ;
    } catch ( illegal_instruction_exception e )
    {
        fprintf ( stderr , "\n\nvm68k: Illegal instruction at 0x%08x\n" ,
                m68kContext -> regs . pc ) ;
    } catch ( zero_divide_exception e )
    {
        fprintf ( stderr , "\n\nvm68k: Divide by zero at 0x%08x\n" ,
                m68kContext -> regs . pc ) ;
    } catch ( privilege_violation_exception e )
    {
        fprintf ( stderr , "\n\nvm68k: Privilege violation at 0x%08x\n" ,
                m68kContext -> regs . pc ) ;
    }
    printRegs () ;
    return 1 ;
}



static void printRegs ( void )
{
    fprintf ( stderr , "registers:\n" ) ;

    for ( int i = 0 ; i < 8 ; i ++ )
    {
        fprintf ( stderr , "\t d%d = 0x%08x\t a%d = 0x%08x\n" ,
                i , m68kContext -> regs . d [ i ] ,
                i , m68kContext -> regs . a [ i ] ) ;
    }
    fprintf ( stderr , "\t pc = 0x%08x\tusp = 0x%08x\n"
                        "\tssp = 0x%08x\n" ,
            m68kContext -> regs . pc ,
            m68kContext -> regs . usp ,
            m68kContext -> regs . ssp ) ;
    fprintf ( stderr , "\tccr = %s %s %s %s %s %s\n" ,
            m68kContext -> regs . ccr . ls () ? "ls" : "hi" ,
            m68kContext -> regs . ccr . cs () ? "cs" : "cc" ,
            m68kContext -> regs . ccr . eq () ? "eq" : "ne" ,
            m68kContext -> regs . ccr . mi () ? "mi" : "pl" ,
            m68kContext -> regs . ccr . lt () ? "lt" : "ge" ,
            m68kContext -> regs . ccr . le () ? "le" : "gt" ) ;
}


