

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>

#include <map>
#include <list>
#include <iostream>

#include "utils.h"
#include "state_machine.h"


/****** State_Machine public methods ******/

/** State_Machine constructor
 *
 * Create a new state machine from the given source file.
 */
State_Machine :: State_Machine ( const char * source_file )
{
    const int       NUMFIELDS = 2 ;
    regmatch_t      matches [ NUMFIELDS + 1 ] ;
    State *         current_state = new State () ;
    FILE *          input = fopen ( source_file , "rt" ) ;
    char            str [ MAX_LINE_LEN + 1 ] ;
    int             line_no = 0 ;
    int             rc ;
    regex_t         statement_with_parameter_regex ;
    regex_t         statement_without_parameter_regex ; 


    /* Compile regular expressions */
    rc = regcomp ( & statement_with_parameter_regex , 
        "^[\t ]*(CALL|JUMP|LABEL) +([A-Za-z0-9_]+)[ \t]*$" ,
            REG_EXTENDED ) ;
    assert ( rc == 0 ) ;

    rc = regcomp ( & statement_without_parameter_regex , 
        "^[\t ]*(CLOCK|RETURN|IDECODE)[ \t]*$" , REG_EXTENDED ) ;
    assert ( rc == 0 ) ;


    /* Begin building the state list */
    states . push_back ( current_state ) ;

    if ( input == 0L )
    {
        throw new File_Access_Exception ( source_file ) ;
    }

    /* Read in lines from the machine definition. */
    while ( fgets ( str , MAX_LINE_LEN , input ) != NULL )
    {
        line_no ++ ;

        Remove_Trailing_Newlines ( str ) ;

        if ( regexec ( & statement_with_parameter_regex , 
                            str , NUMFIELDS + 1 , matches , 0 ) == 0 )
        {
            /* This is one of the state machine statements */
            char *  statement_name = Get_Regex_Match ( str , & matches [ 1 ] ) ;
            char *  parameter_name = Get_Regex_Match ( str , & matches [ 2 ] ) ;

            if ( strcmp ( statement_name , "JUMP" ) == 0 )
            {
                current_state -> Add_Command ( 
                        new Command ( Command :: JUMP , parameter_name ) ) ;
            } else if ( strcmp ( statement_name , "CALL" ) == 0 )
            {
                current_state -> Add_Command ( 
                        new Command ( Command :: CALL , parameter_name ) ) ;
            } else if ( strcmp ( statement_name , "LABEL" ) == 0 )
            {
                /* This state is being assigned a label. */

                char *  label_name = parameter_name ;

                /* We add the label to the label definitions table */
                if ( definitions . count ( string ( label_name ) ) > 0 )
                {
                    throw new Duplicate_Label_Exception ( label_name ) ;
                }

                definitions [ string ( label_name ) ] = current_state ;

                /* For convenience, we also store the state name in the
                 * state itself */
                current_state -> Set_State_Name ( label_name ) ;
            } else { 
                assert ( 0 ) ;
            }
            delete [] statement_name ;
            delete [] parameter_name ;
        } else if ( regexec ( & statement_without_parameter_regex , str , 2 , 
                                            matches , 0 ) == 0 )
        {
            char *  statement_name = Get_Regex_Match ( str , & matches [ 1 ] ) ;

            if ( strcmp ( statement_name , "RETURN" ) == 0 )
            {
                current_state -> Add_Command ( 
                        new Command ( Command :: RETURN ) ) ;
            } else if ( strcmp ( statement_name , "IDECODE" ) == 0 )
            {
                current_state -> Add_Command ( 
                        new Command ( Command :: IDECODE ) ) ;
            } else if ( strcmp ( statement_name , "CLOCK" ) == 0 )
            {
                /* That's the end of state marker. Create a new state. */
                current_state = new State () ;
                /* and add it to the list of states */
                states . push_back ( current_state ) ;
            } else {
                assert ( 0 ) ;
            }
            delete [] statement_name ;
        } else if ( String_Contains_Non_Whitespace ( str ) ) 
        {
            /* This is a VHDL statement or a comment or something
            like that. We don't care precisely what it is. */
            current_state -> Add_Command ( 
                    new Command ( Command :: VHDL , str ) ) ;
        }
    }
    fclose ( input ) ;

    /* A common error is to omit the CLOCK from the last state in the
    machine. We can check for that here: does the last state have
    anything in it? */

    if ( ! current_state -> Is_Empty () )
    {
        throw new Missing_Clock_Exception ( source_file ) ;
    }

    /* Remove last state, since it has no CLOCK */
    states . pop_back () ;
    delete current_state ;

    /* We also check that the machine has some states in it. An empty
    machine is a mistake! */
    if ( states . empty () )
    {
        throw new Empty_Machine_Exception ( source_file ) ;
    }

    /* Store the name of the state machine */
    name = Copy_String ( source_file ) ;

    /* Free the regular expressions */
    regfree ( & statement_without_parameter_regex ) ;
    regfree ( & statement_with_parameter_regex ) ;

    is_finalised = false ;
}

/** State_Machine empty constructor
 *
 * A new empty state machine is created.
 */
State_Machine :: State_Machine ()
{
    is_finalised = false ;
    name = Copy_String ( "unnamed" ) ;
}


/** State_Machine destructor
 *
 * Destroys the state machine and _all_ related memory.
 */
State_Machine :: ~State_Machine ()
{
    /* Free the state list */
    State_List_Iterator iter ;
    State *             item ;
    
    iter = states . begin () ;
    while ( iter != states . end () )
    {
        item = (* iter) ;
        delete item ;
        iter ++ ;
    }

    delete [] name ;
}


/** Incorporate_Sub_Machine
 *
 * The given state machine is incorporated into the current state machine.
 * This is intended to be done in order to bring in other machines required
 * to satisfy dependencies. The sub machine is left empty.
 */
void State_Machine :: Incorporate_Sub_Machine ( State_Machine * sub_machine )
{
    assert ( ! is_finalised ) ;

    /* Move the contents of the sub machine into this one:
     * states and definitions */
    MESSAGE2 ( "Incorporating sub machine %s.\n" , Get_Name () ) ;


    /* Copy the states */
    State_List_Iterator state_list_iter ;
    
    state_list_iter = sub_machine -> states . begin () ;
    while ( state_list_iter != sub_machine -> states . end () )
    {
        State * item = (* state_list_iter) ;

        states . push_back ( item ) ;

        state_list_iter ++ ;
    }

    /* Copy the definitions */
    State_Map_Iterator  state_map_iter ;
    
    state_map_iter = sub_machine -> definitions . begin () ;
    while ( state_map_iter != sub_machine -> definitions . end () )
    {
        const char *    label = (* state_map_iter) . first . c_str () ;
        State *         state = (* state_map_iter) . second ;

        if ( definitions . count ( string ( label ) ) > 0 )
        {
            throw new Duplicate_Label_Exception ( label ) ;
        }
        definitions [ string ( label ) ] = state ;

        state_map_iter ++ ;
    }

    /* Now delete the contents of the sub machine.. This is essential
     * to prevent the memory that makes them up being deleted when
     * the sub machine is deleted. */
    sub_machine -> states . clear () ;
    sub_machine -> definitions . clear () ;
}



/** Compile_Machine
 *
 * Generates the VHDL for the state machine, sending it to the given
 * output device. This can only be done once all the dependencies
 * for the machine, as listed by Depends_On(), have been satisfied.
 * Assertions will fail if dependencies haven't been satisfied.
 */
void State_Machine :: Compile_Machine ( FILE * output )
{
    int                             width_of_state_number ;
    State_List_Iterator             state_list_pos ;
   

    if ( ! is_finalised )
    {
        /* Add absolute state numbers to all states */
        Calculate_Abs_State_Numbers () ;
    }

    width_of_state_number = Get_Width_Of_State_Number () ;

    MESSAGE2 ( "Number of states: %d\n"
                "Width of state number (bits): %d\n" ,
                states . size () , width_of_state_number ) ;

    /* Must be in a process sensitive to 'state' and 'clock' */
    fprintf ( output , 
        "-- Start automatically generated state machine logic.\n"
        "\tcall_requested <= '0' ;\n"
        "\treturn_requested <= '0' ;\n"
        "\tcall_state <= ( others => '0' ) ;\n\n"
        "\tcase state is\n" ) ;

    state_list_pos = states . begin () ;
    while ( state_list_pos != states . end () )
    {
        State * current_state = (* state_list_pos) ;

        current_state -> Generate_VHDL ( output , & definitions ) ;

        state_list_pos ++ ;
    }

    fprintf ( output ,
        "when others => null ;\n"
        "\tend case ;\n\n"
        "-- End automatically generated state machine logic.\n" ) ;
}


/** Depends_On
 *
 * Return a list of states that this machine depends upon.
 * Essentially, this is the set of all states that are JUMPed
 * or CALLed, minus the set of locations within this machine. */
Label_Set State_Machine :: Depends_On ()
{
    Label_Set           depends_on_set ;
    State_List_Iterator state_iter ;
    State *             state ;
    
    state_iter = states . begin () ;
    while ( state_iter != states . end () )
    {
        state = (* state_iter) ;

        Label_Set           state_depends_on_set = state -> Depends_On () ;
        Label_Set_Iterator  label_iter ;

        label_iter = state_depends_on_set . begin () ;
        while ( label_iter != state_depends_on_set . end () )
        {
            const char *    label = (* label_iter) ;

            depends_on_set . insert ( label ) ;

            label_iter ++ ;
        }

        state_iter ++ ;
    }

    /* Now we have the set of all states reached by a JUMP or CALL
     * from this machine. Remove all the states provided by this machine. */

    Label_Set           provides_set = Provides () ;
    Label_Set_Iterator  label_iter ;

    label_iter = provides_set . begin () ;
    while ( label_iter != provides_set . end () )
    {
        const char *    label = (* label_iter) ;

        if ( depends_on_set . count ( label ) != 0 )
        {
            /* This label is depended upon by this machine, but it is
             * also provided by this machine. */
            label_iter ++ ;

            depends_on_set . erase ( label ) ;
        } else {
            label_iter ++ ;
        }
    }

    return depends_on_set ;
}



/** Provides
 *
 * Return the set of labels provided by this machine. This
 * is simply the set of keys of the "definitions" map */
Label_Set State_Machine :: Provides ()
{
    Label_Set           provides_set ;
    State_Map_Iterator  iter ;
    
    iter = definitions . begin () ;
    while ( iter != definitions . end () )
    {
        const char * label = (* iter) . first . c_str () ;

        provides_set . insert ( label ) ;

        iter ++ ;
    }
    
    return provides_set ;
}

/** Finalise_SM
 *
 * The state machine is finalised - no more changes can be made to it.
 */
void State_Machine :: Finalise_SM ()
{
    assert ( ! is_finalised ) ;
    Calculate_Abs_State_Numbers () ;
    is_finalised = true ;
}

/** Get_State_For_Name 
 *
 * Convert a state name into a State object, if possible.
 */
State * State_Machine :: Get_State_For_Name ( const char * sub_name )
{
    if ( definitions . count ( string ( sub_name ) ) > 0 )
    {
        return definitions [ string ( sub_name ) ] ;
    } else {
        return 0L ;
    }
}

/****** State_Machine private methods ******/


/** Calculate_Abs_State_Numbers
 *
 * Work out the absolute state numbers of all the states in the database.
 */
void State_Machine :: Calculate_Abs_State_Numbers ()
{
    MESSAGE2 ( "Calculating absolute state numbers.\n" ) ;

    State_List_Iterator pos ;
    int                 abs_state_number = 0 ;

    pos = states . begin () ;
    while ( pos != states . end () )
    {
        State * state = (* pos) ;
        
        state -> Set_Abs_State_Number ( abs_state_number ) ;

        abs_state_number ++ ;

        pos ++ ;
    }

    /* Program the width of each state number. */
    int                 width_of_state_number = Get_Width_Of_State_Number () ;

    pos = states . begin () ;
    while ( pos != states . end () )
    {
        State * state = (* pos) ;
        
        state -> Set_Width_Of_State_Number ( width_of_state_number ) ;

        abs_state_number ++ ;

        pos ++ ;
    }
}

